/*
 * TopView.cpp
 * 
 * Copyright (C) 2008 Matthias Schiffer <matthias@gamezock.de>
 * 
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "TopView.h"

#include <Instance.h>
#include <Data/Level.h>
#include <Data/Room.h>
#include <Data/Triangle.h>
#include <Gui/RenderArea.h>
#include <Math/Triangle2D.h>

#include <GL/gl.h>
#include <cmath>
#include <set>

namespace ZoomEdit {
namespace View {

bool TopView::Edge::operator<(const Edge &e) const {
  if(v1->x() < e.v1->x())
    return true;
  if(v1->x() > e.v1->x())
    return false;
  if(v1->y() < e.v1->y())
    return true;
  if(v1->y() > e.v1->y())
    return false;
  if(v1->z() < e.v1->z())
    return true;
  if(v1->z() > e.v1->z())
    return false;
  if(v2->x() < e.v2->x())
    return true;
  if(v2->x() > e.v2->x())
    return false;
  if(v2->y() < e.v2->y())
    return true;
  if(v2->y() > e.v2->y())
    return false;
  if(v2->z() < e.v2->z())
    return true;
  if(v2->z() > e.v2->z())
    return false;
  
  return false;
}

void TopView::drawGrid() {
  float depth = 1.25f + 0.04f*zoomLevel;
  float depth2 = std::floor(depth);
  float step = std::pow(0.1f, depth2);
  float width = viewWidth/scale;
  float height = viewHeight/scale;
  
  float x1 = xCenter - width/2;
  float x2 = xCenter + width/2;
  float y1 = yCenter - height/2;
  float y2 = yCenter + height/2;
  
  glLineWidth(1.0f);
  
  glBegin(GL_LINES);
  
  for(int i = 0; 0.4f*(depth-depth2+i-1) < 0.5f; ++i) {
    float f = std::min(0.4f*(depth-depth2+i), 0.5f);
    glColor3f(f, f, f);
    
    for(f = x1 - std::fmod(x1, step); f <= x2; f+=step) {
      glVertex2f(f, y1);
      glVertex2f(f, y2);
    }
    
    for(f = y1 - std::fmod(y1, step); f <= y2; f+=step) {
      glVertex2f(x1, f);
      glVertex2f(x2, f);
    }
    
    step *= 10;
  }
  
  glEnd();
}

void TopView::renderRoom(Data::Room *room, bool selected) {
  const std::list<Data::Triangle*> &triangles = room->getTriangles();
  std::multiset<Edge> edges;
  
  if(selected)
    glColor4f(0.0f, 0.7f, 1.0f, 0.2f);
  else
    glColor4f(0.0f, 0.7f, 1.0f, 0.3f);
  
  glBegin(GL_TRIANGLES);
  
  for(std::list<Data::Triangle*>::const_iterator t = triangles.begin(); t != triangles.end(); ++t) {
    vmml::vec3f edge1 = (*t)->getVertex(1) - (*t)->getVertex(0);
    vmml::vec3f edge2 = (*t)->getVertex(2) - (*t)->getVertex(0);

    vmml::vec3f normal = edge1.cross(edge2);

    if(normal.y() <= 0)
      continue;

    for(int i = 0; i < 3; ++i) {
      const vmml::vec3f &v = (*t)->getVertex(i);
      const vmml::vec3f &v2 = (*t)->getVertex((i+1)%3);
      
      glVertex2f(v.x(), v.z());
      
      edges.insert(Edge(&v, &v2));
    }
  }
  
  glEnd();
  
  for(std::multiset<Edge>::iterator next = edges.begin(); next != edges.end();) {
    std::multiset<Edge>::iterator edge = next++;
    std::multiset<Edge>::iterator edge2 = edges.find(Edge(edge->v2, edge->v1));
    
    if(edge2 != edges.end()) {
      edges.erase(edge2);
      next = edge;
      ++next;
      edges.erase(edge);
    }
  }
  
  if(selected) {
    glColor4f(1.0f, 1.0f, 1.0f, 0.9f);
    glLineWidth(2.0f);
  }
  else {
    glColor4f(0.0f, 0.7f, 1.0f, 0.7f);
    glLineWidth(1.0f);
  }
  
  glBegin(GL_LINES);
  
  for(std::multiset<Edge>::iterator edge = edges.begin(); edge != edges.end(); ++edge) {
    glVertex2f(edge->v1->x(), edge->v1->z());
    glVertex2f(edge->v2->x(), edge->v2->z());
  }
  
  glEnd();
}

void TopView::init() {
  glClearColor(0, 0, 0, 0);
  
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  glEnable(GL_BLEND);
  
  glEnable(GL_LINE_SMOOTH);
  glEnable(GL_POINT_SMOOTH);
}

void TopView::resize(float width, float height) {
  if(width == viewWidth && height == viewHeight)
    return;
  
  viewWidth = width; viewHeight = height;
  
  glViewport(0, 0, width, height);
  
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  
  if(width != 0 && height != 0)
    glScalef(2.0f/width, -2.0f/height, 1);
  
  signalUpdate().emit();
}

void TopView::zoom(int zoom, float x, float y) {
  const float oldScale = scale;
  
  zoomLevel = std::max(std::min(zoomLevel + zoom, 50), -100);
  scale = 100*std::pow(1.1f, zoomLevel);
  
  xCenter += x*(1/oldScale - 1/scale);
  yCenter += y*(1/oldScale - 1/scale);
  
  signalUpdate().emit();
}

void TopView::move(float x, float y, unsigned int state) {
  if(!(state & GDK_BUTTON3_MASK))
    return;
  
  xCenter += x/scale;
  yCenter += y/scale;
  
  signalUpdate().emit();
}

void TopView::render() {
  glClear(GL_COLOR_BUFFER_BIT);
  
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  
  glScalef(scale, scale, 1);
  glTranslatef(-xCenter, -yCenter, 0);
  
  drawGrid();
  
  if(!instance->getLevel())
    return;
  
  const std::list<Data::Room*> &rooms = instance->getLevel()->getRooms();
  
  for(std::list<Data::Room*>::const_iterator room = rooms.begin(); room != rooms.end(); ++room)
    renderRoom(*room, *room == instance->getSelectedRoom());
}

void TopView::click(float x, float y, unsigned int button) {
  if(button != 1)
    return;
  
  vmml::vec2f v(xCenter + x/scale, yCenter + y/scale);
  
  const std::list<Data::Room*> &rooms = instance->getLevel()->getRooms();
  
  for(std::list<Data::Room*>::const_iterator room = rooms.begin(); room != rooms.end(); ++room) {
    const std::list<Data::Triangle*>& triangles = (*room)->getTriangles();
    
    for(std::list<Data::Triangle*>::const_iterator t = triangles.begin(); t != triangles.end(); ++t) {
      vmml::vec3f edge1 = (*t)->getVertex(1) - (*t)->getVertex(0);
      vmml::vec3f edge2 = (*t)->getVertex(2) - (*t)->getVertex(0);

      vmml::vec3f normal = edge1.cross(edge2);

      if(normal.y() <= 0)
        continue;

      vmml::vec2f v1((*t)->getVertex(0).x(), (*t)->getVertex(0).z());
      vmml::vec2f v2((*t)->getVertex(1).x(), (*t)->getVertex(1).z());
      vmml::vec2f v3((*t)->getVertex(2).x(), (*t)->getVertex(2).z());
      
      if(Math::Triangle2D(v1, v2, v3).contains(v)) {
        instance->setSelectedRoom(*room);
        return;
      }
    }
  }

  instance->setSelectedRoom(0);
}
 
}
}
